import sympy as sp
import unittest
from IntegralElimination import *

class TestIntegralElimination(unittest.TestCase):
    def test_08(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta")
        M = IM(1,x(t)) 
        N = IM(1,y(t))
        IA = IntegralAlgebra(order=[x(t),y(t)],
                            parameters=[theta])
        
        MN = IA.monomial_product(M,N).get_sympy_repr()
        
        MN_expected = IM(1, x(t), y(t)) + IM(1, y(t), x(t))
        verify = sp.simplify(MN_expected - MN) == 0
        self.assertTrue(verify)

if __name__ == '__main__':
    unittest.main()


import sympy as sp
import unittest
from IntegralElimination import *

class TestIntegralElimination(unittest.TestCase):
    def test_01(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        test_OK = False
        try:
            IM(x(t)+y(t))
        except AssertionError:
            test_OK = True
        self.assertTrue(test_OK)


if __name__ == '__main__':
    unittest.main()


from .IntegralMonomial import IM

def check_condition_LM_half_reduced_product(IA, P, N): 
    """ 
    Following the definition of the half-reduced product,
    (case where N != 1)
    if (P_{I}⌋1+⌉ · N) > (N⌋1+⌉ · P_{N}) then
    lm{P half-reduced-prod N) = lm{ int ( PI⌋1+⌉ · N ) }  
    
    This function returns True if the condition holds.
    Future work should investigate the other cases.
    """
    if N == IM(1): return True
    lm_PI = IA.LM_LC(P.get_P_I())[0]
    lm_PN = IA.LM_LC(P.get_P_N())[0]

    expr1 = IA.LM_LC(IA.monomial_product(lm_PI.cut("1+"),N))[0]
    expr2 = IA.LM_LC(IA.monomial_product(N.cut("1+"),lm_PN))[0] 

    return  not IA.IMO_le(expr1,expr2)


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .IntegralAlgebra import IntegralAlgebra
 
import sympy as sp
from ordered_set import OrderedSet

from .IntegralMonomial import IM
from .IntegralPolynomial import IntegralPolynomial 
from .utils import (
    expr_has_symbols
)

def find_A_A0_G_F(IA: IntegralAlgebra,
                    P:IntegralPolynomial) -> tuple[IntegralPolynomial]:
    """
    This algorithm 
    takes an integral polynomial P
    and tries to write it in the form
    P = A - (AA_0 + int{ A · G + F }$
    where 
    A_0 is a constant, 
    A,G,F are integral polynomials,
    A and G are nonzero.
    """
    #try to write P as A − (A0 +int Q)
    P_N = P.get_P_N()
    CST = P.get_cst_terms()
    A = IA.polynomials_subtraction(P_N ,CST)
    A0 = IA.product_P_Coeff(CST,-1)
    P_I = P.get_P_I()
    Q = IA.product_P_Coeff(P_I.cut_P("1+"),-1) 

    # check that A != 0
    if A.is_zero(): return None
    # and check that A is in K[X]  
    if any([M.get_nb_int() for M,_ in A.get_content()]): 
        return None
    
    #convert A to simple polynomial ie sp.Expr
    A_sp = sp.Add(*[M.get_content()[0]*coeff for M,coeff in A.get_content()])
     
    r = [] 

    for M_i, alpha_i in Q.get_content():
        M0 = M_i.cut('0').get_content()[0] 
        q_i, r_i = sp.reduced(M0, [A_sp], *IA.used_order, order="lex") 
        q_i = q_i[0]#we use only one pol to reduce so we select qi0
        q_i_mons_coeff = list(q_i.as_coefficients_dict(IA.t).items())
        #return [(m0,coeff0), (m1,coeff1),...]
        r_i_mons_coeff = list(r_i.as_coefficients_dict(IA.t).items())
        q_i_pol = IntegralPolynomial(
            {IM(mons): coeff for mons,coeff in q_i_mons_coeff} 
        )
        r_i_pol = IntegralPolynomial(
            {IM(mons): coeff for mons,coeff in r_i_mons_coeff}  
        )
        r += [(alpha_i, q_i_pol, r_i_pol, M_i)]
        
    G, F = IntegralPolynomial(0), IntegralPolynomial(0)

    # this loop is used to create G and F 
    # G = sum (alpha_i q_i M_i[i1+])
    # F = sum (alpha_i r_i M_i[i1+]) 
    for alpha_i, q_i, r_i, M_i in r: 
        M_i_cut = IntegralPolynomial(M_i.cut("i1+"))
        M_i_cut_alpha = IA.product_P_Coeff(M_i_cut,alpha_i)
        temp_G = IA.polynomials_product(q_i,M_i_cut_alpha) 
        G = IA.polynomials_add(G,temp_G)
        temp_F = IA.polynomials_product(r_i,M_i_cut_alpha) 
        F = IA.polynomials_add(F,temp_F)
    
    if G.is_zero():
        return None
    
    # check that P = A-(A0+\int (AG+F))
    # going from right to left :
    temp = IA.polynomials_product(A,G)
    temp = IA.integrate_polynomial(IA.polynomials_add(temp,F))
    temp = IA.polynomials_add(A0, temp)
    temp = IA.polynomials_subtraction(A,temp)
    assert sp.simplify(P.get_sympy_repr() - temp.get_sympy_repr()) == 0
    return (A,A0,G,F)




def update_exp(IA: IntegralAlgebra,
                T_prime: OrderedSet[IntegralPolynomial],
                E: OrderedSet[sp.Function, sp.Function, IntegralPolynomial]
                ) -> tuple:
    """
    When find_A_A0_G_F succeeds,
    a new polynomial A - (AA_0 u + u int{ vF }) can be introduced
    where u and v are new indeterminates encoding 
    u = e^{int{G}} and 
    v = e^{-int{G}}.
    """
    T_E = OrderedSet()
    E_prime = OrderedSet([elem for elem in E])
    t = IA.t 
    for P in T_prime:
        temp = IA.find_A_A0_G_F(P)
        if temp != None:
            A, A0, G, F = temp
            int_G = IA.integrate_polynomial(G)

            # this allow to not introduce an encoding of an exp
            # that is already in E_prime 
            int_G = IA.reduce(int_G, T_prime)[0] 
            
            # check that int G != from all Q_i in Eprime
            check = [
                    IA.polynomials_subtraction(int_G, Q_i).is_zero() 
                    for _, _, Q_i in E_prime
                    ]

            if F.is_zero():
                pass #divide instead of exp ?
            
            if not any(check):
                k = len(E_prime) + 1
                uk = sp.Function(f"u{k}")
                vk = sp.Function(f"v{k}")
                new_item = (uk(t), vk(t), int_G)
                E_prime.add(new_item)
                uk_pol = IntegralPolynomial(IM(uk(t)))
                vk_pol = IntegralPolynomial(IM(vk(t)))
            else: 
                index_True_in_check = [ 
                                        i for i in range(len(check)) 
                                        if check[i] is True
                                      ][0]
                uk,vk,_ = list(E_prime)[index_True_in_check]
                uk_pol = IntegralPolynomial(IM(uk))
                vk_pol = IntegralPolynomial(IM(vk))
            # then, we compute P_exp = A - (A0 u_k + u_k \int (v_k F))
            # from right to left 
            temp = IA.integrate_polynomial(IA.polynomials_product(vk_pol, F))
            temp = IA.polynomials_product(uk_pol,temp)
            temp = IA.polynomials_add(IA.polynomials_product(A0,uk_pol), temp)
            P_exp = IA.polynomials_subtraction(A, temp)
            if not P_exp.is_zero():
                T_E.add(P_exp)

    return T_E, E_prime


def update_log(IA: IntegralAlgebra,
                T_prime: OrderedSet[IntegralPolynomial],
                L: OrderedSet[sp.Function, IntegralPolynomial]
                ) -> tuple:
    """
    When find_A_A0_G_F succeeds,
    a new polynomial l-int{G} can be introduced
    where l is a new indeterminates encoding 
    l = ln{A/A_0}.
    """
    T_L = OrderedSet()
    L_prime = OrderedSet([elem for elem in L])
    t = IA.t 
    for P in T_prime:
        temp = IA.find_A_A0_G_F(P)
        if temp != None:
            A, A0, G, F = temp 
            if F.is_zero():
                assert len(A0.get_content()) == 1
                A0_val = A0.get_content()[0][1]
                A_div_A0 = IA.product_P_Coeff(A,1/A0_val)
 
                # this allow to not introduce an encoding of an exp
                # that is already in E_prime 
                A_div_A0_red = IA.reduce(A_div_A0, T_prime)[0] 
                
                # check that A != from all Q_i in Lprime
                check = [
                        IA.polynomials_subtraction(A_div_A0_red, Q_i).is_zero() 
                        for _, Q_i in L_prime
                        ]
                
                
                if not any(check):
                    k = len(L_prime) + 1
                    lk = sp.Function(f"l{k}") 
                    new_item = (lk(t), A_div_A0_red)
                    L_prime.add(new_item)
                    lk_pol = IntegralPolynomial(IM(lk(t))) 
                else: 
                    index_True_in_check = [ 
                                            i for i in range(len(check)) 
                                            if check[i] is True
                                        ][0]
                    lk,_ = list(L_prime)[index_True_in_check]
                    lk_pol = IntegralPolynomial(IM(lk)) 
                # then, we compute 
                # P_log  =  ln(A/A0) - int(G)
                # from right to left
                int_G = IA.integrate_polynomial(G)
                P_log = IA.polynomials_subtraction(
                    lk_pol,
                    int_G
                )
                
                if not P_log.is_zero():
                    T_L.add(P_log)

    return T_L, L_prime
 

def extend_X_with_exp_and_log(IA: IntegralAlgebra,
                    E: OrderedSet[sp.Function, sp.Function, IntegralPolynomial],
                    L: OrderedSet[sp.Function, IntegralPolynomial]
                    ) -> list:
    """
    Algorithm 5
    Here, X is inside the IntegralAlgebra object (self.order)
    We will return a modified X and replace the self.used_order 
    after this function call to update self.IMO_le according to
    the new X'
    """ 
    X_prime = IA.order
    EL = E | L 
    all_funcs = set()
    for elem in EL:
        all_funcs = all_funcs | {*elem[:-1]}

    i=1 
    while len(EL) > 0:
        elem = EL.pop(0)
        #elem is (Func,Func,Pol) for exp or (Func, Pol) for log
        new_funcs, Qi = elem[:-1], elem[-1]
        if expr_has_symbols(Qi.get_sympy_repr(), X_prime):
            indeterminates = Qi.get_time_dependant_functions()
            greatest_inds = None

            #from left (greatest) to right
            for inds, k in zip(X_prime, range(len(X_prime))): 
                if inds in indeterminates:
                    greatest_inds = inds
                    X_prime = X_prime[:k] + [*new_funcs] + X_prime[k:]
                    break;
            assert greatest_inds is not None
        else:
            if expr_has_symbols(Qi.get_sympy_repr(), all_funcs):
                EL.add(elem)
            else:
                X_prime = X_prime + [*new_funcs]
 
        if i == 1000000:
            raise RuntimeError("infinite loop: cyclic graph")
        i+=1
    return X_prime


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .IntegralAlgebra import IntegralAlgebra

from ordered_set import OrderedSet

from .IntegralPolynomial import IntegralPolynomial 


def integral_elimination(IA: IntegralAlgebra,
                        F: OrderedSet[IntegralPolynomial],
                        disable_exp: bool = False,
                        disable_log: bool = True,
                        disable_critical_pairs: bool = False,
                        nb_iter = None) -> tuple:

    T = OrderedSet([elem for elem in F])
    E = OrderedSet()
    L = OrderedSet()
    X_prime = IA.order
    finished = False 
    
    i = 1
    while not finished:
        T_prime = IA.auto_reduce(T)
        E_prime = E
        L_prime = L
        T_E = OrderedSet([])
        T_L = OrderedSet([])
        if not disable_exp:
            T_E, E_prime = IA.update_exp(T_prime, E)
        if not disable_log:
            T_L, L_prime = IA.update_log(T_prime, L)
        X_prime = IA.extend_X_with_exp_and_log(E_prime, L_prime)
        IA.update_IMO(X_prime)
        T_prime = T_prime | T_E | T_L # T_prime union T_E union T_L

        if not disable_critical_pairs:
            C = IA.critical_pairs(T_prime)
            for Q in C:
                # _ is a boolean = has_been_reduced
                Q_red, _ = IA.reduce(Q,T_prime) 
                if not Q_red.is_zero():
                    T_prime.add(Q_red)

        E_prime_red = OrderedSet([])
        for ui, vi, Qi in E_prime:
            # _ is a boolean = has_been_reduced
            Qi_red, _ = IA.reduce(Qi, T_prime) 
            E_prime_red.add((ui, vi ,Qi_red))

        L_prime_red = OrderedSet([])
        for li, Qi in L_prime:
            # _ is a boolean = has_been_reduced
            Qi_red, _ = IA.reduce(Qi, T_prime) 
            L_prime_red.add((li ,Qi_red))

        X_prime = IA.extend_X_with_exp_and_log(E_prime_red, L_prime_red)

        IA.update_IMO(X_prime)
        E = E_prime_red 
        L = L_prime_red 
        
        T_equal_T_prime = check_T_equal_T_prime(T, T_prime)
        T = T_prime

        if (i == nb_iter) or T_equal_T_prime:
            finished = True

        i += 1
    return (E, L, T, X_prime)


def check_T_equal_T_prime(T: OrderedSet[IntegralPolynomial],
                          T_prime: OrderedSet[IntegralPolynomial]) -> bool:
    T_sp = OrderedSet([eq.get_sympy_repr() for eq in T])
    T_prime_sp = OrderedSet([eq.get_sympy_repr() for eq in T_prime]) 
    return T_sp == T_prime_sp


import sympy as sp
from ordered_set import OrderedSet

from .utils import expr_has_symbol
from .IntegralMonomial import IM
from .IntegralPolynomial import IntegralPolynomial

def add_alpha_M(d, alpha, M):
        if M in d:
            new_coeff = sp.cancel(alpha + d[M])
            if new_coeff !=0:
                d[M] = new_coeff
            else:
                del d[M]
        else:
            d[M] = alpha

def first_order_ODE_to_IntegralPolynomial(
                                        expr:sp.Expr, 
                                        var=sp.Symbol("t"),
                                        use_symbol=False
                                        ) -> IntegralPolynomial:
        expr = sp.expand(expr)
        dict_coeff = dict(expr.as_coefficients_dict(var))
        res = {}
        if expr_has_symbol(expr, sp.Integral): raise ValueError
        expr_has_der = expr_has_symbol(expr,sp.Derivative)
        for mons, coeff in dict_coeff.items():
            if mons.func == sp.Derivative : 
                assert mons.args[1][1] == 1 
                m = mons.args[0]
                if use_symbol:
                    IC = sp.Symbol(f"{m.func}0")
                else:
                    IC = m.func(0)
                M = IM(m)
                add_alpha_M(res,coeff,M)
                add_alpha_M(res,-IC,IM(1)) 
            else:
                if expr_has_der:
                    M = IM(1,mons) 
                    add_alpha_M(res,coeff,M)
                else:
                    M = IM(mons)
                    add_alpha_M(res,coeff,M)
        return IntegralPolynomial(res)

def ODE_sys_to_Integral_sys(sys: list[sp.Expr]
                            ) -> OrderedSet[IntegralPolynomial]:
    sys = OrderedSet([first_order_ODE_to_IntegralPolynomial(eq) for eq in sys])
    return  sys


import sympy as sp
import unittest
from IntegralElimination import *

class TestIntegralElimination(unittest.TestCase):
    def test_07(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta")
        P = IntegralPolynomial(3*IM(x(t))+theta*IM(y(t),1,y(t))+theta*IM(1,1,y(t))) 
        
         
        P_N = P.get_P_N().get_sympy_repr()
        P_I = P.get_P_I().get_sympy_repr()
 
        expected_P_N = 3*IM(x(t))+theta*IM(y(t),1,y(t))
        expected_P_I = theta*IM(1,1,y(t)) 

        verifiy_P_N = sp.simplify(expected_P_N-P_N) == 0
        verifiy_P_I = sp.simplify(expected_P_I-P_I) == 0
        
        self.assertTrue(verifiy_P_N and verifiy_P_I)

if __name__ == '__main__':
    unittest.main()


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .IntegralAlgebra import IntegralAlgebra

import sympy as sp 
from ordered_set import OrderedSet
from itertools import combinations

from .IntegralMonomial import IM
from .IntegralPolynomial import IntegralPolynomial 
from .utils import (
    expr_has_symbol,
    expr_has_symbols
)

from .check_conditions_LM import (
    check_condition_LM_half_reduced_product,
)


def critical_pairs_PI_QI(A: IntegralAlgebra, 
                         P: IntegralPolynomial,
                         Q: IntegralPolynomial) -> IntegralPolynomial:
    """
    Let P, Q such that
	P_I, P_N, Q_I, Q_N are nonzero
	and
	lm{P} = lm{P_I}, lm{Q} = lm{Q_I}.
	
	Let alpha, beta positive integers
	and 
    B_U = lm{P^{reduced power {alpha}}}[2+],
	B_V = lm{Q^{reduced power {beta}}}[2+].
	%
	If
	lm{P^{reduced power {alpha}}}[1] = lm{Q^{reduced power {beta}}}[1]
	and if the condition of 
    check_condition_LM_half_reduced_product
    (this function checks the condition)
    holds for
	both 
    P^{reduced power {alpha}} half-reduced-product  B_V
        (this expression is denoted U)
    and 
    Q^{reduced power {beta}} half-reduced-product  B_U 
        (this expression is denoted V),
	then  define

    S(P,Q) = (1/l_U) U - (1/l_V) V 
    
	where  l_U (resp. l_V) is the leading coefficient of
	U (resp. V).
    """  
    if P.is_zero() or Q.is_zero():
        return None
    #first, we check that lm(P)=lm(PI) and same for Q
    LM_P,LC_P = A.LM_LC(P)
    LM_Q,LC_Q = A.LM_LC(Q)
    assert LC_P == 1, "LC(P) should be normalized"
    assert LC_Q == 1, "LC(Q) should be normalized"
    P_I = P.get_P_I()
    Q_I = Q.get_P_I()
    if P_I.is_zero() or Q_I.is_zero():
        return None
    LM_P_I,LC_P_I = A.LM_LC(P_I)
    LM_Q_I,LC_Q_I = A.LM_LC(Q_I)
    if (LM_P_I != LM_P) or (LM_Q_I != LM_Q):
        return None
    

    # now, we have to find if there exist 
    # positives integers alpha and beta such that 
    # lm(P_I)[1]*lm(P_N)[0]**(alpha-1) = lm(Q_I)[1]*lm(Q_N)[0]**(beta-1)
    
    # first, we have to create all the variables
    alpha = sp.Symbol("alphaaa") # to avoid having alpha if the user use it 
    beta = sp.Symbol("betaaa")
    P_N = P.get_P_N()
    Q_N = Q.get_P_N()
    if P_N.is_zero() or Q_N.is_zero():
        return None
    LM_P_N,LC_P_N = A.LM_LC(P_N)
    LM_Q_N,LC_Q_N = A.LM_LC(Q_N)
    LM_P_I_1 = LM_P_I.cut("1").get_content()[0] # lm(P_I)[1]
    LM_Q_I_1 = LM_Q_I.cut("1").get_content()[0] # lm(Q_I)[1]
    LM_P_N_0 = LM_P_N.cut("0").get_content()[0]# lm(P_N)[0]
    LM_Q_N_0 = LM_Q_N.cut("0").get_content()[0]# lm(P_N)[0]
    # let's try to solve the equation to find alpha and beta 
    eq = (LM_P_I_1*LM_P_N_0**(alpha-1))/(LM_Q_I_1*LM_Q_N_0**(beta-1))  
    # we search alpha and beta such that eq = 1
    eq_pow = eq.as_powers_dict() 
    if len(eq_pow) != 2:
        return None 
    for k,v in eq_pow.items():
        # exclude the cases such as LM_PI = \int \int C, LM_PN = C
        # LM_QI = \int I, LM_QN = C 
        # we search alpha beta such that 
        # C^(alpha-1) = C^(beta-1) I
        # but it's not possible
        if not expr_has_symbols(v, [alpha,beta]):
            return None
    solved_pows = sp.solve(list(eq_pow.values())) 
    alpha_value = solved_pows[alpha] 
    beta_value = solved_pows[beta] 

    if alpha_value < 1 or beta_value < 1:
        return None

    # we verify that 
    # lm(P_I)[1]*lm(P_N)[0]**(alpha-1) = lm(Q_I)[1]*lm(Q_N)[0]**(beta-1)
    sanity_check = sp.simplify(
                            LM_P_I_1*LM_P_N_0**(alpha_value-1) -\
                            (LM_Q_I_1*LM_Q_N_0**(beta_value-1)) 
                            ) == 0
    assert sanity_check


    # we first compute the reduced power of P and Q
    P_circled_alpha = A.reduced_power(P,alpha_value) 
    Q_circled_beta  = A.reduced_power(Q, beta_value)

    # we compute BV and BU
    # 1) B_V:
    #---------
    LM_Q_I_i2plus = IntegralPolynomial(LM_Q_I.cut("i2+")) # lm(Q_I)[i2+] 
    # we need to convert LM_Q_N_i1plus to Int pol to use power
    LM_Q_N_i1plus = IntegralPolynomial(LM_Q_N.cut("i1+")) 
    LM_Q_N_i1plus_pow = A.polynomial_power(LM_Q_N_i1plus, beta_value-1)
    B_V = A.LM_LC(A.polynomials_product(LM_Q_I_i2plus, LM_Q_N_i1plus_pow))[0]
    # 2) B_U:
    #---------
    LM_P_I_i2plus = IntegralPolynomial(LM_P_I.cut("i2+")) # lm(P_I)[i2+]
    LM_P_N_i1plus = IntegralPolynomial(LM_P_N.cut("i1+"))
    LM_P_N_i1plus_pow = A.polynomial_power(LM_P_N_i1plus, alpha_value-1)
    B_U = A.LM_LC(A.polynomials_product(LM_P_I_i2plus, LM_P_N_i1plus_pow))[0]

 
    # we verify the condition on BV and BU
    if not check_condition_LM_half_reduced_product(A,
        P_circled_alpha, B_V):
        return None
    if not check_condition_LM_half_reduced_product(A,
        Q_circled_beta, B_U):
        return None
 

    # we can now compute S(P,Q)
    # we start by computing U
    U = A.half_reduced_product(P_circled_alpha, B_V)

    # we divide by the coeff of lm(U): 
    _,LC_U = A.LM_LC(U)
    U = A.product_P_Coeff(U, 1/(LC_U))

    # Now we compute V 
    V = A.half_reduced_product(Q_circled_beta, B_U)
 
    # we divide by the coeff of lm(V): 
    _,LC_V= A.LM_LC(V)
    V = A.product_P_Coeff(V, -1/(LC_V)) 
    
    #check that lm(U) == lm(V)
    LM_U,LC_U = A.LM_LC(U)
    LM_V,LC_V = A.LM_LC(V)
    #we verify that lc_U = lc_V BUT since we already multiplied 
    #lc_V by minus 1, we need to check that lc_U = -lc_V ! 
    assert (LM_U == LM_V) and (LC_U + LC_V ==0)

    # We can compute S(P,Q):
    S = A.polynomials_add(U,V)

    # if S(P,Q) != 0 assert that lm(S(P,Q)) < lm(U)
    if not S.is_zero():
        assert A.IMO_le(A.LM_LC(S)[0], A.LM_LC(U)[0])
    return S










def critical_pairs_PI_QN(A: IntegralAlgebra, 
                         P: IntegralPolynomial,
                         Q: IntegralPolynomial) -> IntegralPolynomial:
    """
    Let P, Q such that
	P_I, P_N, Q_I, Q_N are nonzero
	and
	lm{P} = lm{P_I}, lm{Q} = lm{Q_N}.
	Let alpha, beta positive integers,
	a monomial m and
    B_U = lm{P^{reduced power {alpha}}}[2+],
    B_V = lm{Q}[1+].
	If
	$lm{P^{reduced power {alpha}}}[1] =
		m lm{Q_N}[0]$
	and if the condition of 
    check_condition_LM_half_reduced_product
    (this function checks the condition)
    holds for 
	P^{reduced power {alpha}} half-reduced-product  B_V
        (this expression is denoted U),
	then define

    S(P,Q) = (1/l_U) U - (1/l_V) V 
    
	where  
    V = int{m Q · B_U}
    and
    l_U (resp. l_V) is the leading coefficient of
	U (resp. V).
    """
    if P.is_zero() or Q.is_zero():
        return None
    
    #first, we check that lm(P)=lm(PI) and lm(Q) = lm(PN)

    is_LM_in_P_I = A.is_LM_in_P_I(P)
    is_LM_in_Q_N = A.is_LM_in_P_N(Q)
    if is_LM_in_P_I != is_LM_in_Q_N:
        return None #case P_I Q_I or P_N Q_n
    elif not is_LM_in_P_I and not is_LM_in_Q_N:
        # we need to swap p and q
        P,Q = Q,P

    _,LC_P = A.LM_LC(P)
    _,LC_Q = A.LM_LC(Q)
    assert LC_P == 1, "LC(P) should be normalized"
    assert LC_Q == 1, "LC(Q) should be normalized"
    LM_P_I, _ = A.LM_LC(P.get_P_I())
    LM_Q_N, _ = A.LM_LC(Q.get_P_N())

    # now, we have to find if there exist 
    # a positive integers alpha and a monomial (not integral) m such that
    # lm(P_I)[1] lm(PN)[0]**(alpha-1) = m lm(QN)[0]

    # first, we have to create all the variables
    alpha = sp.Symbol("alphaaa") # to avoid having alpha if the user use it 
    P_N = P.get_P_N()
    if P_N.is_zero():
        return None
    LM_P_N,_ = A.LM_LC(P_N) 

    LM_P_I_1 = LM_P_I.cut("1").get_content()[0] # lm(P_I)[1]
    LM_P_N_0 = LM_P_N.cut("0").get_content()[0]# lm(P_N)[0]
    LM_Q_N_0 = LM_Q_N.cut("0").get_content()[0]# lm(P_N)[0]
    
    # let's try to solve the equation to find alpha and m
    # we will devide lm(P_I)[1] lm(PN)[0]**(alpha-1) by lm(QN)[0]
    # the power dict will give one or two things: 
    # 1. only an alpha dependent term, wich mean m will 
    # be 1 after solving alpha
    # 2. one term in alpha an the other will be like {"x(t):2"},
    # then we will pick m = x(t) to cancel it.
    eq = (LM_P_I_1*LM_P_N_0**(alpha-1))/(LM_Q_N_0)  
    # we search alpha  such that eq = 1 
    eq_pow = eq.as_powers_dict()  
    if len(eq_pow) == 1:
        m=1
        alpha_value = sp.solve(list(eq_pow.values()))[alpha]
    elif len(eq_pow) == 2:
        for k,v in eq_pow.items():
            if expr_has_symbol(v, alpha):
                alpha_value = sp.solve(v)[0]
            else:
                # we don't want m to be 1/y(t) (for example)
                if v == -1:
                    return None
                m = k**v
    else:
        return None
     
    if alpha_value < 1:
        return None

    # we verify that 
    # lm(P_I)[1] lm(PN)[0]**(alpha-1) = m lm(QN)[0]
    sanity_check = sp.simplify(
                            (LM_P_I_1*LM_P_N_0**(alpha_value-1)) - \
                            (m*LM_Q_N_0)
                            ) == 0
    assert sanity_check

    # we first compute the reduced power of P 
    P_circled_alpha = A.reduced_power(P,alpha_value) 

    # we compute B_U and B_V
    # 1) B_V
    B_V = LM_Q_N.cut("i1+") 
    # 2) B_U
    LM_P_I_i2plus = IntegralPolynomial(LM_P_I.cut("i2+")) # lm(P_I)[i2+]
    LM_P_N_i1plus = IntegralPolynomial(LM_P_N.cut("i1+"))
    LM_P_N_i1plus_pow = A.polynomial_power(LM_P_N_i1plus, alpha_value-1)
    B_U = A.LM_LC(A.polynomials_product(LM_P_I_i2plus, LM_P_N_i1plus_pow))[0]


    # we verify the condition on BV 
    if not check_condition_LM_half_reduced_product(A,
        P_circled_alpha, B_V):
        return None

    # we can now compute S(P,Q)
    # we start by computing U 
    U = A.half_reduced_product(P_circled_alpha, B_V)
    # we divide by the coeff of lm(U): 
    _,LC_U = A.LM_LC(U)
    U = A.product_P_Coeff(U, 1/(LC_U))

    # Now we compute V
    m = IM(m) # to use the monomial product
    m_B_U = A.monomial_product(m,B_U)
    V = A.integrate_polynomial(A.polynomials_product(Q,m_B_U))
    # we divide by the coeff of lm(V): 
    _,LC_V= A.LM_LC(V)
    V = A.product_P_Coeff(V, 1/(LC_V)) 
     
    #check that lm(U) == lm(V)
    LM_U,LC_U = A.LM_LC(U)
    LM_V,LC_V = A.LM_LC(V)
    #we verify that lc_U = lc_V  
    assert (LM_U == LM_V) 
    assert (LC_U - LC_V ==0)

    # We can compute S(P,Q):
    S = A.polynomials_subtraction(U,V)

    # if S(P,Q) != 0 assert that lm(S(P,Q)) < lm(U)
    if not S.is_zero():
        assert A.IMO_le(A.LM_LC(S)[0], A.LM_LC(U)[0])
    return S










def critical_pairs_PN_QN(A: IntegralAlgebra, 
                         P: IntegralPolynomial,
                         Q: IntegralPolynomial) -> IntegralPolynomial:
    """
    Let P, Q such that  
	lm{P} = lm{P_N}, lm{Q} = lm{Q_N}.
	We denote by L the least common multiple (LCM) of
	lm{P}[0] and lm{Q}[0],
	i.e.  $L=LCM(lm{P}[0], lm{Q}[0])$.

    1) if  lm{Q_N}[1+] != lm{P_N}[1+]
        U = (L/lm{P_N}[0]) · P · lm{Q_N}[1+] 
        V = (L/lm{Q_N}[0]) · Q · lm{P_N}[1+]


    2) if  \LtQdot[1+]=\LtPdot[1+]
        U = (L/lm{P_N}[0]) · P 
        V = (L/lm{Q_N}[0]) · Q 

    Let us define:

    S(P,Q) = (1/l_U) U - (1/l_V) V  
     
    where  
    l_U (resp. l_V) is the leading coefficient of
	U (resp. V).
    """
    if P.is_zero() or Q.is_zero():
        return None
    LM_P,LC_P = A.LM_LC(P)
    LM_Q,LC_Q = A.LM_LC(Q)
    assert LC_P == 1, "LC(P) should be normalized"
    assert LC_Q == 1, "LC(Q) should be normalized"

    #first, we check that lm(P)=lm(PN) and lm(Q) = lm(PN)
    P_N = P.get_P_N()
    Q_N = Q.get_P_N()
    if P_N.is_zero() or Q_N.is_zero():
        return None
    LM_P_N,LC_P_N = A.LM_LC(P_N)
    LM_Q_N,LC_Q_N = A.LM_LC(Q_N)
    if (LM_P_N != LM_P) or (LM_Q_N != LM_Q):
        return None
    
    # we first compute L = LCM(lm(P_N)[0], lm(Q_N)[0])
    LM_P_N_0 = LM_P_N.cut("0").get_content()[0]
    LM_Q_N_0 = LM_Q_N.cut("0").get_content()[0]
    L = sp.lcm(LM_P_N_0,LM_Q_N_0)
    if L == LM_P_N_0*LM_Q_N_0:
        return None
    # we have to determine in which case we are :
    # it means we must determine if lm(QN)[i1+] equal lm(PN)[i1+] or not
    LM_P_N_i1plus = LM_P_N.cut("i1+")
    LM_Q_N_i1plus = LM_Q_N.cut("i1+")
    # before that, to optimize, we can already 
    # compute the common part of U and V
    l_P = LC_P_N
    mons_U = IntegralPolynomial(IM(L/LM_P_N_0))
    U = A.product_P_Coeff(A.polynomials_product(mons_U, P), 1/(l_P))
    l_Q = LC_Q_N
    mons_V = IntegralPolynomial(IM(L/LM_Q_N_0))
    V = A.product_P_Coeff(A.polynomials_product(mons_V, Q), 1/(l_Q))
    if LM_P_N_i1plus != LM_Q_N_i1plus:
        LM_P_N_i1plus = IntegralPolynomial(LM_P_N_i1plus)
        LM_Q_N_i1plus = IntegralPolynomial(LM_Q_N_i1plus)
        U = A.polynomials_product(U,LM_Q_N_i1plus)
        V = A.polynomials_product(V,LM_P_N_i1plus)
    # we divide by the coeff of lm(U): 
    _,LC_U = A.LM_LC(U)
    U = A.product_P_Coeff(U, 1/(LC_U))
    # we divide by the coeff of lm(V): 
    _,LC_V= A.LM_LC(V)
    V = A.product_P_Coeff(V, 1/(LC_V)) 
    #check that lm(U) == lm(V)
    LM_U,LC_U = A.LM_LC(U)
    LM_V,LC_V = A.LM_LC(V)
    #we verify that lc_U = lc_V   
    assert (LM_U == LM_V) and (LC_U - LC_V ==0)

    # We can compute S(P,Q):
    S = A.polynomials_subtraction(U,V)

    # if S(P,Q) != 0 assert that lm(S(P,Q)) < lm(U)
    if not S.is_zero():
        assert A.IMO_le(A.LM_LC(S)[0], A.LM_LC(U)[0])
    return S










def critical_pairs(A: IntegralAlgebra, 
                   R: OrderedSet[IntegralPolynomial]
                ) -> OrderedSet[IntegralPolynomial]:
    """
    see Algorithm 2
    """
    S = OrderedSet()
    
    all_distinct_pairs_in_R = list(combinations(R, 2)) 
    
    for P,Q in all_distinct_pairs_in_R:
        P_norm = A.normalize_LC_of_P(P)[0]
        Q_norm = A.normalize_LC_of_P(Q)[0]
        S_P_Q = A.critical_pairs_PI_QI(P_norm,Q_norm)
        if S_P_Q is not None:
            S.append(S_P_Q)
        S_P_Q = A.critical_pairs_PI_QN(P_norm,Q_norm)
        if S_P_Q is not None:
            S.append(S_P_Q)
        S_P_Q = A.critical_pairs_PN_QN(P_norm,Q_norm)
        if S_P_Q is not None:
            S.append(S_P_Q)

    return S
        

 




import sympy as sp
import unittest
from IntegralElimination import *

class TestIntegralElimination(unittest.TestCase):
    def test_04(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta")
        P = IntegralPolynomial(3*IM(x(t))+theta*IM(y(t),1,y(t))+theta*IM(1,1,y(t)))
        IA = IntegralAlgebra(order=[x(t),y(t)],
                            parameters=[theta])
        
        LM, LC = IA.LM_LC(P) 
        self.assertEqual((LM, LC), (IM(x(t)),3))


if __name__ == '__main__':
    unittest.main()


import sympy as sp
import unittest
from IntegralElimination import *

class TestIntegralElimination(unittest.TestCase):
    def test_05(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta")
        P_expected = 3*IM(x(t))+theta*IM(y(t),1,y(t))+theta*IM(1,1,y(t))
        P = IntegralPolynomial(P_expected)                
        P_sp= P.get_sympy_repr()
        equal = sp.simplify(P_sp - P_expected) == 0
        self.assertTrue(equal)

if __name__ == '__main__':
    unittest.main()


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .IntegralAlgebra import IntegralAlgebra

import sympy as sp 
from ordered_set import OrderedSet

from .IntegralMonomial import IM
from .IntegralPolynomial import IntegralPolynomial 
from .utils import (
    is_int, 
)

from .check_conditions_LM import (
    check_condition_LM_half_reduced_product,
)


def reduction_M_by_P_simple_case(A: IntegralAlgebra, 
                                 M: IM, 
                                 P: IntegralPolynomial) -> IntegralPolynomial:
    """
    Consider an integral monomial M
    and a nonzero integral polynomiaml P.
	
	If there exists an integral monomial 
    N such that lm{lm{P} · N}=M,
	then introduce
	R =  M-(1/l)P · N
	where l is the leading coefficient of P · N.
    """
    assert not P.is_zero() 

    LM_P,LC_P = A.LM_LC(P)
    assert LC_P == 1, "LC(P) should be normalized"

    N = A.monomial_division(M,LM_P)
    if N == None:
        return None  
    
    PN = A.polynomials_product(P,IntegralPolynomial(N))
    PN = A.normalize_LC_of_P(PN)[0]
    
    assert M == A.LM_LC(PN)[0]

    R = A.polynomials_subtraction(
            IntegralPolynomial(M),
            PN
        ) 
    if not R.is_zero():
        LM_R = A.LM_LC(R)[0]
        assert A.IMO_le(LM_R,M)
    return R
 

def reduction_M_by_P_reduced_power(A: IntegralAlgebra, 
                                   M: IM, 
                                   P: IntegralPolynomial
                                   ) -> IntegralPolynomial:
    """
    Consider an integral monomial M and an integral polynomial P
    with P_I != 0, P_N != 0, and lm{P} = lm{P_I}$.
    
    Assume there exists a positive integer n
    and an integral monomial N such that
    - lm{P_I}[1] lm{lm{P_N}[0]}^{n-1} = M[1],
    - lm{ (P_I[2+]) · (P_N[1+])^{n-1} · N } = M[2+],
    - N[0] = 1.

    If the condition
    of check_condition_LM_half_reduced_product
    (this function checks the condition)
    holds for P^{\reduced_power{n}} half-reduced-prod N,  
    then introduce
	R = M - (1/l) M[0](P^{reduced_power{n}} half-reduced-prod N )
	where
	l is the leading coefficient 
    of M[0](P^{reduced_power{n}} half-reduced-prod N ).
    """
    assert not P.is_zero() 
    LM_P,LC_P = A.LM_LC(P)
    assert LC_P == 1, "LC(P) should be normalized"
    if M.get_nb_int() == 0 : # |M| >= 1
        return None
    P_I = P.get_P_I()
    if P_I.is_zero(): return None 
    LM_P_I,LC_P_I = A.LM_LC(P_I)
    if LM_P_I != LM_P : # we want lm(P) = lm(P_I)
        return None
    P_N = P.get_P_N()
    if P_N.is_zero(): return None 
    LM_P_N,LC_P_N = A.LM_LC(P_N)
    M_1 = M.cut("1").get_content()[0]
    LM_P_N_0 = LM_P_N.cut("0").get_content()[0]
    LM_P_I_1 = LM_P_I.cut("1").get_content()[0]

    # we try to find n such that (lm(PI)[1]*lm(PN)[0]**(n-1))/M[1] = 1
    n = sp.Symbol("n_pow")
    pow_dict = ((LM_P_I_1*LM_P_N_0**(n-1))/M_1).as_powers_dict()
    if len(pow_dict) > 1: 
        #for example :
        # expr = ((x(t)*a(t))**n*y(t))/(x(t)**2*y(t))
        # expr.as_powers_dict()
        # {a(t)x(t):n, x(t): -2}
        return None 
    sp_solved = sp.solve(list(pow_dict.values())[0])
    if len(sp_solved)==0: return None
    solved_n = sp_solved[0] 
    if not(is_int(solved_n) and solved_n > 0):
        return None
    #we then have to verify the condition:
    #lm( (lm(P_I)[2+]) cdot (lm(P_N)[1+])**(n-1)) = M[2+]
    LM_P_I_i2plus = IntegralPolynomial(LM_P_I.cut("i2+"))
    LM_P_N_i1plus = IntegralPolynomial(LM_P_N.cut("i1+"))
    LM_P_N_i1plus_pow = A.polynomial_power(LM_P_N_i1plus,solved_n-1)
    pol = A.polynomials_product(LM_P_I_i2plus, LM_P_N_i1plus_pow)
    
    Mi2_plus = M.cut("i2+")
    B_P, c_B_P = A.LM_LC(pol)
    N = A.monomial_division(Mi2_plus, B_P)
    if N is None or N.get_content()[0] != 1:
        return None
    
    # we compute the reduced power
    P_reduced_pow_n = A.reduced_power(P,solved_n)
    # we verify the condition on N
    if not check_condition_LM_half_reduced_product(A, P_reduced_pow_n, N):
        return None
    # we can now compute R
    P_reduced_pow_n_red_prod_N = A.half_reduced_product(P_reduced_pow_n, N)
    
    M0 = IntegralPolynomial(M.cut("0"))
    M0_P_pow_n = A.polynomials_product(P_reduced_pow_n_red_prod_N, M0)
    l_P = LC_P_N
    temp = A.product_P_Coeff(M0_P_pow_n, -1/(solved_n*l_P**(solved_n-1)))
    assert M == A.LM_LC(temp)[0]
    R = A.polynomials_add(IntegralPolynomial(M), temp)
    if not R.is_zero():
        LM_R = A.LM_LC(R)[0]
        assert A.IMO_le(LM_R,M)
    return R 

def reduction_M_by_P(A: IntegralAlgebra, 
                     M: IM, 
                     P: IntegralPolynomial
                    ) -> IntegralPolynomial:
    """
    Lemma 14
    """
    assert not P.is_zero() 
    e = M.get_nb_int() 
    for i in reversed(range(e+1)): 
        prefix = M.get_prefix(i)
        suffix = M.get_suffix(i)   
        R = A.reduction_M_by_P_simple_case(suffix,P)  
        if R is not None:
            R = A.add_prefix_to_polynomial(prefix,R) 
            if not R.is_zero():
                LM_R = A.LM_LC(R)[0]
                assert A.IMO_le(LM_R,M)
            return R
        R = A.reduction_M_by_P_reduced_power(suffix,P) 
        if R is not None:
            R = A.add_prefix_to_polynomial(prefix,R)
            if not R.is_zero():
                LM_R = A.LM_LC(R)[0]
                assert A.IMO_le(LM_R,M)
            return R 

def reduce(IA:IntegralAlgebra, 
           Q:IntegralPolynomial, 
           T:OrderedSet[IntegralPolynomial],
           has_been_reduced=False) -> IntegralPolynomial:
    """
    reduce Q by the set of integral polynomials T
    """ 
    if Q.is_zero(): 
        return IntegralPolynomial(0), False
    A = Q 

    LM_A,LC_A = IA.LM_LC(A)
    #test if LM can be reduced by some P of T
    if IA.storage["P_red_to_zero"].get(A):
        return IntegralPolynomial(0), True
    for P in T: 
        P_norm, _ = IA.normalize_LC_of_P(P)  
        R = reduction_M_by_P(IA, LM_A, P_norm)  
        if R is not None:  
            #A=lm(A)+B, i.e B is the tail, so we can just delete the leading 
            # monomial of A to obtain B
            B = A.get_content_as_dict()
            del B[LM_A]

            alphaR = IA.product_P_Coeff(R, LC_A, return_dict=True)
            A = IA.polynomials_add(B, alphaR, return_dict=False)   
            if A.is_zero():
                return A, True 
            return reduce(IA, A, T, has_been_reduced=True) 
    if A.is_zero():
        IA.storage["P_red_to_zero"][A] = IntegralPolynomial(0)
    return A, has_been_reduced

def __auto_reduce(IA:IntegralAlgebra, 
                T:OrderedSet[IntegralPolynomial]) -> tuple:
    T_reduced =  OrderedSet([])
    T_done = OrderedSet([]) 
    one_P_has_been_reduced = False
    for P in T:
        T_done.add(P) 
        T_copy = T - T_done | T_reduced 
        P_reduced, has_been_reduced = IA.reduce(P,T_copy)
        one_P_has_been_reduced = one_P_has_been_reduced or has_been_reduced 
        if not P_reduced.is_zero(): 
            T_reduced.add(P_reduced) 
        
    return T_reduced, one_P_has_been_reduced

def auto_reduce(IA:IntegralAlgebra, 
                T:OrderedSet[IntegralPolynomial]
            ) -> OrderedSet[IntegralPolynomial]:
    T_reduced = T
    has_been_reduced = True 
    while has_been_reduced:  
        T_temp, has_been_reduced = __auto_reduce(IA,T_reduced) 
        T_reduced = T_temp  
    return T_reduced


import sympy as sp
import unittest
from IntegralElimination import *
from ordered_set import OrderedSet

class TestIntegralElimination(unittest.TestCase):
    def test_16(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta") 

        x0,y0 = sp.symbols("x0 y0")
        theta = sp.Symbol("theta")
        A = IntegralAlgebra(order=[x(t),y(t)],
                            parameters=[theta,x0,y0])
        P1 = IntegralPolynomial(IM(x(t))-x0*IM(1) - IM(1,y(t)))
        P2 = IntegralPolynomial(IM(y(t))-y0*IM(1) - IM(1,x(t)*y(t)))
        T = OrderedSet([P1,P2]) 
        T_red = A.auto_reduce(T)
        T_red = set([P.get_sympy_repr() for P in T_red])
        expected_sys = {-x0*IM(1) + IM(x(t)) - IM(1, y(t)), -x0*IM(1, y(t)) - y0*IM(1) + IM(y(t)) - IM(1, y(t), y(t))}
        verify = len(T_red - expected_sys) == 0
        self.assertTrue(verify)

if __name__ == '__main__':
    unittest.main()


import sympy as sp
import unittest
from IntegralElimination import *

class TestIntegralElimination(unittest.TestCase):
    def test_11(self):
        x = sp.Function('x')
        y = sp.Function('y')
        t = sp.Symbol("t")
        theta = sp.Symbol("theta") 

        IA = IntegralAlgebra(order=[x(t),y(t)],
                            parameters=[theta])
        
        P = IntegralPolynomial(IM(1,x(t))-IM(y(t)))
        reduce_product= IA.reduced_product(P,IM(1,y(t))).get_sympy_repr()
        
        expected = IM(1, y(t)**2) - IM(y(t), y(t)) + IM(1, x(t), y(t))
        verify = sp.simplify(expected - reduce_product) == 0
        self.assertTrue(verify)

if __name__ == '__main__':
    unittest.main()


import sympy as sp
from functools import lru_cache

from .utils import (
    has_add_in_list,
    is_float
)

def has_multiplicative_constant(expr): 
    """
    Search if expr has a multiplicative cst != than 1
    """ 
    if is_float(expr):
        if expr == 1:
            return False
        else:
            return True
    elif expr.is_Mul:
        factors = expr.as_ordered_factors() 
        for factor in factors:
            if factor.is_number and factor != 1:
                return True  
    return False


class IM(sp.Function):
    @lru_cache(None)  # Cache unlimited number of calls
    def __new__(cls, *args):
        obj = super(IM, cls).__new__(cls, *args)
        obj.c = args
        obj.t = sp.Symbol("t")
        assert "IM" not in str(obj.c)
        assert not has_add_in_list(obj.c)

        #check if parameters (symbols) are in self.c
        assert not any([
                        len(elem.free_symbols - {obj.t}) 
                        for elem in obj.c if not is_float(elem)
                       ])
        #check that every multiplicative csts are 1
        assert not any([
                        has_multiplicative_constant(elem) 
                        for elem in obj.c
                       ])
        return obj

    def get_nb_int(self):
        return len(self.c) - 1

    def get_content(self):
        return self.c
    
    def cut(self, cut_type):
        """
        See Definition 7 
        """
        if cut_type ==  "0":
            return IM(self.c[0])
        
        elif cut_type ==  "1":
            assert self.get_nb_int() >= 1
            return IM(self.c[1])
        
        elif cut_type ==  "1+":
            assert self.get_nb_int() >= 1
            return IM(*self.c[1:])
        
        elif cut_type ==  "i1+":
            if self.get_nb_int() == 0:
                return IM(1)
            else:
                return IM(1,*self.c[1:])
          
        elif cut_type == "i2+":
            if self.get_nb_int() == 1: 
                return IM(1)
            else:
                assert self.get_nb_int() >= 2
                return IM(1,*self.c[2:])
        else:
            raise NotImplementedError
        
    def get_suffix(self, i):
        assert 0 <= i <= self.get_nb_int()
        return IM(*self.get_content()[i:])
    
    def get_prefix(self, i):
        assert 0 <= i <= self.get_nb_int()
        return IM(*self.get_content()[:i])

    def add_prefix(self, M):
        prefixed_M = IM(*M.get_content(),*self.get_content())
        return prefixed_M

    def __str__(self):
        return f"IM({str(self.c)[1:-1]})"
    
    def get_integral_repr(self): 
        if len(self.c) > 1:
            expr = sp.Integral(self.c[-1], self.t)
            for e in reversed(self.c[1:-1]): 
                expr = (sp.Integral(e*expr, self.t))
            expr = self.c[0]*expr
        elif len(self.c) == 1:
            expr = self.c[0]
        return expr

    def _latex(self, printer=None):  
        return f"{printer.doprint(self.get_integral_repr())}"
    



import sympy as sp
from ordered_set import OrderedSet


from .IntegralPolynomial import IntegralPolynomial
from .utils import expr_has_symbols


def get_IO_equations(T: OrderedSet[IntegralPolynomial],
                    X_prime: list[sp.Function],
                    unknowns: list[sp.Function]):
        """
        unknowns are the unknowns function without the exponentials
        example : 
        T = {
        IntegralPolynomial(-x0*y0*IM(1, u1(t)*u2(t)) - y0*IM(1) + y0*IM(u2(t)))
        IntegralPolynomial(-x0*IM(u1(t)) + IM(x(t)))
        IntegralPolynomial(-y0*IM(u2(t)) + IM(y(t)))
        IntegralPolynomial(x0*IM(1, u1(t)) + x0*IM(1)/theta - x0*IM(u1(t))/theta)
        }
        X_prime = [x(t), y(t), u2(t), v2(t), u1(t), v1(t)]
        unknowns = [x(t)]

        return :
        I = {
        IntegralPolynomial(-x0*y0*IM(1, u1(t)*u2(t)) - y0*IM(1) + y0*IM(u2(t)))
        IntegralPolynomial(-y0*IM(u2(t)) + IM(y(t)))
        IntegralPolynomial(x0*IM(1, u1(t)) + x0*IM(1)/theta - x0*IM(u1(t))/theta)
        }
        """
        i = 0
        M = X_prime[i]
        while (i < len(X_prime)) and (M != unknowns[-1]):
            i+=1
            M = X_prime[i]
        idx_last_unk = i 
        
        new_unknows = X_prime[:idx_last_unk+1]

        I = []
        for eq in T:
            if not expr_has_symbols(eq.get_sympy_repr(), new_unknows):
                I += [eq]
        return I

def convert_IO_equations(T)->OrderedSet[sp.Expr]:
    """
    convert a list of IntegralPolynomial to integral repr if needed
    """
    T_conv = []
    for eq in T: 
        if isinstance(eq, IntegralPolynomial):
            T_conv += [eq.get_integral_repr()]
        else:
            T_conv += [eq]
    return T_conv

def replace_exp_in_IO_eqs(
        E: OrderedSet[sp.Function, sp.Function, IntegralPolynomial],
        I: list[IntegralPolynomial]) -> list[sp.Expr]:
    """
    example: 
    E = {
        (u1(t), v1(t), IntegralPolynomial(theta*IM(1, 1)))
        (u2(t), v2(t), IntegralPolynomial(-x0*IM(1)/theta + x0*IM(u1(t))/theta))
    }
    I = {
        IntegralPolynomial(-x0*y0*IM(1, u1(t)*u2(t)) - y0*IM(1) + y0*IM(u2(t)))
        IntegralPolynomial(-y0*IM(u2(t)) + IM(y(t)))
        IntegralPolynomial(x0*IM(1, u1(t)) + x0*IM(1)/theta - x0*IM(u1(t))/theta)
    }

    return : 
    I_exp = {
        -x0*y0*Integral(exp(theta*Integral(1, t))*exp(x0*exp(theta*Integral(1, t))/theta - x0/theta), t) + y0*exp(x0*exp(theta*Integral(1, t))/theta - x0/theta) - y0
        -y0*exp(x0*exp(theta*Integral(1, t))/theta - x0/theta) + y(t)
        x0*Integral(exp(theta*Integral(1, t)), t) - x0*exp(theta*Integral(1, t))/theta + x0/theta
    }
    """
    I_exp = []
    I = convert_IO_equations(I)
    for eq in I: 
        eq_subs = eq
        for (u,v,IntG) in reversed(E):
            IntG = IntG.get_integral_repr()
        
            eq_subs = eq_subs.subs({u: sp.exp(IntG),
                                    v: sp.exp(-IntG)})
        I_exp += [eq_subs]
    return I_exp


def replace_ln_in_IO_eqs(
        L: OrderedSet[sp.Function, IntegralPolynomial],
        I: list[IntegralPolynomial]) -> list[sp.Expr]: 
    L_ln = []
    I = convert_IO_equations(I)
    for eq in I: 
        eq_subs = eq
        for (li, AdivA0) in reversed(L):
            AdivA0 = AdivA0.get_integral_repr()
        
            eq_subs = eq_subs.subs({li: sp.ln(AdivA0)})
        L_ln += [eq_subs]
    return L_ln

def change_integral_bounds(expr: sp.Expr,
                           bound):
    """
    bound can be a symbol or a tuple
    example: change sp.Integral(x(t),t) to sp.Integral(x(t),(t,0,t))
            the inverse is also possible using bound = t
    """ 
    if len(expr.args) == 0:
        return expr
    elif expr.func == sp.Integral:
        new_arg = change_integral_bounds(expr.args[0], bound)
        return sp.Integral(new_arg, bound)
    else:
        new_args = [change_integral_bounds(arg,bound) for arg in expr.args] 
        new_fn = expr.func(*new_args)
        return new_fn


def simplify_IO_equations(I: list[sp.Expr],
                        order: list[sp.Function],
                        parameters: list[sp.Symbol]) -> list[sp.Expr]:
    
    t = sp.Symbol("t")
    I_bounds  = [change_integral_bounds(eq, (t,0,t)) for eq in I]
    params_subs = {
        p: sp.Symbol(str(p), nonzero=True, real=True) 
            for p in parameters
        }
    ci_subs = { 
        sp.Symbol(f"{m.func}0"): sp.Symbol(f"{m.func}0", 
                                           nonzero=True, 
                                           real=True) 
        for m in order
        }
    subs_dict = params_subs | ci_subs
    I_simp = [sp.simplify(eq.subs(subs_dict).doit()) for eq in I_bounds]
    I_simp = [change_integral_bounds(eq, t) for eq in I_simp]
    return I_simp


import sympy as sp
from ordered_set import OrderedSet
from IPython.display import display, Math

from .ordering import IMO, comp_lexico, comp_lexico_lists
from .IntegralMonomial import IM
from .utils import (
    is_int,
    shuffle_list,
    subtract_lists,
    lyndon_decomp,
    dicts_addition,
    dicts_subtraction,
    dict_mul_by_coeff,
    expr_has_symbol
)
from .IntegralPolynomial import IntegralPolynomial
from .reduction import (
    reduction_M_by_P_simple_case,
    reduction_M_by_P_reduced_power, 
    reduction_M_by_P, 
    reduce,auto_reduce
)
from .critical_pairs import (
    critical_pairs_PI_QI,
    critical_pairs_PI_QN,
    critical_pairs_PN_QN,
    critical_pairs
)
from .exp_log import (
    find_A_A0_G_F, 
    update_exp, 
    update_log,
    extend_X_with_exp_and_log
) 
from .integral_elimination import integral_elimination

class IntegralAlgebra():
    """
    integral polynomials will be represented as follows :
    eq = {IM(x(t):3, IM(y(t),1,y(t)):theta+lambda}
    using eq.as_coefficients_dict(IM)
    It means that eq is the sum of the two tuples 
    """
    def __init__(self, order, parameters):
        self.order = order
        self.t = sp.Symbol("t")
        self.used_order = order
        self.used_order_str = str(order)
        self.IMO_le = lambda m1, m2 : IMO(m1, m2, self.used_order) 
        self.parameters = parameters
        self.storage = self.init_storage()

    def init_storage(self):
        storage = {
            "LM_LC": {},
            "P_red_to_zero": {}
        }
        return storage
    
    def update_IMO(self, order):
        self.used_order = order
        self.used_order_str = str(order)
        self.IMO = lambda m1, m2 : IMO(m1, m2, self.used_order)
 
    def LM_LC(self, P:IntegralPolynomial) -> tuple[IM, sp.Expr]:
        """
        get leading monomials and coeff
        """
        if P.is_zero():
            return None
        str_order = self.used_order_str
        if not self.storage["LM_LC"].get(str_order):
            self.storage["LM_LC"][str_order]={}
        if not self.storage["LM_LC"][str_order].get(P): 
            L = P.get_content() 
            LC, LM = L[0][1], L[0][0]
            for M,coeff in L:   
                #if True then M <= LM, else M > LM
                if self.IMO_le(M, LM) == False: 
                    LM = M # im > LM
                    LC = coeff
            self.storage["LM_LC"][str_order][P] = (LM,LC)
        LM, LC = self.storage["LM_LC"][str_order][P]
        return LM, LC
    
    def LM_LC_P_I(self, P:IntegralPolynomial) -> tuple[IM, sp.Expr]:
        if P.is_zero():
            return 
        P_I = P.get_P_I()
        if P_I.is_zero(): return None
        return self.LM_LC(P_I)
    
    def LM_LC_P_N(self, P:IntegralPolynomial) -> tuple[IM, sp.Expr]:
        if P.is_zero():
            return 
        P_N = P.get_P_N()
        if P_N.is_zero(): return None
        return self.LM_LC(P_N)

    def is_LM_in_P_I(self, P:IntegralPolynomial) ->  bool:
        LM_P, _ = self.LM_LC(P)
        P_I = P.get_P_I()
        if P_I.is_zero(): return False
        LM_P_I, _ = self.LM_LC(P_I)
        return LM_P_I == LM_P
    
    def is_LM_in_P_N(self, P:IntegralPolynomial) ->  bool:
        LM_P, _ = self.LM_LC(P)
        P_N = P.get_P_N()
        if P_N.is_zero(): return False
        LM_P_N, _ = self.LM_LC(P_N)
        return LM_P_N == LM_P
        
    def display_rules(self, sys):
        display(Math(r"\{")) 
        for eq in sys:
            LM, LC = self.LM_LC(eq)
            key = IntegralPolynomial({LM:LC})
            value = self.polynomials_subtraction(eq, key)
            value = self.product_P_Coeff(value,-1)
            arrow = r"{\Huge \color{red} \mathbf{\rightarrow}}"  
            key_repr = key.repr_display_math()
            value_repr = value.repr_display_math()
            export = Math(f"{key_repr} {arrow} {value_repr}") 
            display(export)
        display(Math(r"\}"))

    def display_rules_without_coeff(self, sys):
        display(Math(r"\{")) 
        for eq in sys:
            LM, LC = self.LM_LC(eq)
            key = IntegralPolynomial({LM:1})
            value = eq.get_content_as_dict()
            del value[LM]
            for k,v in value.items():
                value[k]=1
            value=IntegralPolynomial(value)
            # value = self.product_P_Coeff(value,-1)
            arrow = r"{\Huge \color{red} \mathbf{\rightarrow}}"  
            key_repr = key.repr_display_math()
            value_repr = value.repr_display_math()
            export = Math(f"{key_repr} {arrow} {value_repr}") 
            display(export)
            print(self.parameters)
            print([expr_has_symbol(eq.get_integral_repr(),s) for s in self.parameters])
    
        display(Math(r"\}"))
         
    def display_LMs(self, sys):
        display(Math(r"\{")) 
        for eq in sys:
            LM, LC = self.LM_LC(eq)
            key = IntegralPolynomial({LM:LC})
            arrow = r"{\Huge \color{red} \mathbf{\rightarrow}}"  
            key_repr = key.repr_display_math()
            export = Math(f"{key_repr} {arrow} ...") 
            display(export)
        display(Math(r"\}"))
         

    def normalize_LC_of_P(self, 
                        P: IntegralPolynomial,
                        simplify_coeff=True) -> IntegralPolynomial:
        _, LC = self.LM_LC(P)
        normalized_P = {}
        for M, coeff in P.get_content(): 
            normalized_P[M] = coeff/LC
        P_norm = IntegralPolynomial(normalized_P, simplify_coeff=simplify_coeff)
        return P_norm, LC
    
    def monomial_product(self, 
                          M: IM, 
                          N: IM,
                          simplify_coeff=True)-> IntegralPolynomial:
        e = M.get_nb_int()
        f = N.get_nb_int()
        m0n0 = M.get_content()[0]*N.get_content()[0]
        if e == f == 0:
            MN = {IM(m0n0):1}
        elif e == 0 and f != 0:
            N1p = N.cut("1+").get_content() #N1plus
            MN  = {IM(m0n0,*N1p):1} 
        elif e != 0 and f== 0 :
            M1p = M.cut("1+").get_content() #N1plus
            MN = {IM(m0n0,*M1p): 1}
        else:
            N1p = N.cut("1+").get_content() #N1plus
            M1p = M.cut("1+").get_content() #N1plus
            sh = shuffle_list(M1p,N1p)
            mons_dict = {}
            for elem in sh:
                mons = IM(m0n0,*elem)
                if mons not in mons_dict:
                    mons_dict[mons] =1
                else:
                    mons_dict[mons] += 1 
            MN = mons_dict
        return IntegralPolynomial(MN,simplify_coeff=simplify_coeff)


    def monomial_division(self, M, T):
        """
        find an integral monomial N such that lm(T \cdot N) = M
        i.e. we "divide" M by T.

        return N if N exists. N=1 if M=T
        return None otherwise
        """
        if T.get_nb_int() > M.get_nb_int():
            return None
        if not self.IMO_le(T,M):
            return None
        T_c = T.get_content()
        M_c = M.get_content()
        if T_c[0] == M_c[0] == 1:
            N_0 = 1
        else:
            q, r = sp.div(M_c[0],T_c[0])
            if r != 0: return None
            N_0 = q 
        T_c, M_c = T_c[1:], M_c[1:]
        T_lyn_rev = lyndon_decomp(w      = list(reversed(T_c)), 
                                  cmp_fn = comp_lexico, 
                                  order  = self.used_order)  
        M_lyn_rev = lyndon_decomp(w      = list(reversed(M_c)), 
                                  cmp_fn = comp_lexico, 
                                  order  = self.used_order)  

        N_lyn_rev= subtract_lists(M_lyn_rev, T_lyn_rev)

        if T_lyn_rev != M_lyn_rev and len(N_lyn_rev)==0:
            return None
        N = [N_0]
        for elem in reversed(N_lyn_rev):
            N += [*reversed(elem)]
        N = IM(*N)
        
        #check that lm(T \cdot N) = M
        T_dot_N = self.monomial_product(T, N)
        LM_T_dot_N, _ = self.LM_LC(T_dot_N)
        assert LM_T_dot_N == M
        return N 
    
    def product_P_Coeff(self, 
                        P: {IntegralPolynomial, dict}, 
                        alpha: sp.Expr,
                        simplify_coeff=True,
                        return_dict=False) -> {IntegralPolynomial, dict}:
        """
        alpha is cst 
        """
        if type(P) == dict:
            alpha_P = dict_mul_by_coeff(P, alpha)
        elif type(P) == IntegralPolynomial:
            alpha_P = dict_mul_by_coeff(P.get_content_as_dict(), alpha)
        else:
            raise ValueError
        if return_dict:
            return alpha_P
        return IntegralPolynomial(alpha_P,simplify_coeff=simplify_coeff)

    def polynomials_add(self, 
                        P: {IntegralPolynomial, dict}, 
                        Q: {IntegralPolynomial, dict},
                        simplify_coeff=True,
                        return_dict=False) -> {IntegralPolynomial, dict}: 
        if type(P) == dict and type(Q) == dict:
            L = [P,Q]
        elif type(P) == IntegralPolynomial and type(Q) == IntegralPolynomial:
            L = [P.get_content_as_dict(), Q.get_content_as_dict()]
        else:
            raise ValueError
        PplusQ = dicts_addition(L)
        if return_dict:
            return PplusQ
        return IntegralPolynomial(PplusQ,simplify_coeff=simplify_coeff)
    
    def polynomials_subtraction(self, 
                                P: IntegralPolynomial, 
                                Q: IntegralPolynomial,
                                simplify_coeff=True,
                                return_dict=False) -> {IntegralPolynomial, dict}: 
        if type(P)==dict and type(Q)==dict:
            L = [P,Q]
        elif type(P)==IntegralPolynomial and type(Q)==IntegralPolynomial:
            L = [P.get_content_as_dict(), Q.get_content_as_dict()]
        else:
            raise ValueError
        PminusQ = dicts_subtraction(L)
        if return_dict:
            return PminusQ
        return IntegralPolynomial(PminusQ,simplify_coeff=simplify_coeff)
    

    def polynomials_product(self,
                            P: IntegralPolynomial,
                            Q: IntegralPolynomial,
                            simplify_coeff=True) -> IntegralPolynomial: 
        PQ = IntegralPolynomial(0)
        for M_P, c_P in P.get_content():
            for M_Q, c_Q in Q.get_content(): 
                M_PdotM_Q = self.monomial_product(
                                M_P,
                                M_Q,
                                simplify_coeff=simplify_coeff
                            )
                c_P_c_Q_M_PdotM_Q = self.product_P_Coeff(
                                        M_PdotM_Q,
                                        c_P*c_Q,
                                        simplify_coeff=simplify_coeff
                            )
                PQ = self.polynomials_add(PQ, 
                                          c_P_c_Q_M_PdotM_Q,
                                        simplify_coeff=simplify_coeff
                            )
        return PQ 
    
    def integrate_monomial(self, M: IM):
        return IM(1,*M.get_content())

    def integrate_polynomial(self, 
                             P: IntegralPolynomial,
                             simplify_coeff=False) -> IntegralPolynomial:
        Int_P = {}
        for M, coeff in P.get_content():
            Int_M = self.integrate_monomial(M)
            Int_P[Int_M] = coeff
        return IntegralPolynomial(Int_P,simplify_coeff=simplify_coeff)
 
    def add_prefix_to_polynomial(self, 
                                prefix: IM, 
                                P: IntegralPolynomial) -> IntegralPolynomial:
        new_P = {} 
        for M, coeff in P.get_content():
            pref_M = M.add_prefix(prefix)
            new_P[pref_M] = coeff
        return IntegralPolynomial(new_P)

    def half_reduced_product(self, 
                        P:IntegralPolynomial, 
                        N:IM,
                        simplify_coeff=True) -> IntegralPolynomial:
        """
        Let N an integral monomial and P an integral polynomial
        such that 
        - N[0]=1, 
        - P_I != 0,
        - lm{P} = lm{P_I}.
        1) If N=1:
        half-reduced-product = P 

        2) Otherwise,
        half-reduced-product =   
            P · N - int{ N]1+] · P }
            
        or equivalently 
        
        half-reduced-product = 
            int{ (P_I]1+] · N) - (N]1+] · P_N) } + P_N · N        
        """   
        assert (N.cut("0") == IM(1))
        if N == IM(1): return P

        LM_P,_ = self.LM_LC(P)
        P_I = P.get_P_I()
        assert not P_I .is_zero()
        LM_P_I,_ = self.LM_LC(P_I)
        assert LM_P == LM_P_I 
        N1p = IntegralPolynomial(N.cut("1+"),simplify_coeff=simplify_coeff)
        N = IntegralPolynomial(N,simplify_coeff=simplify_coeff)
        #(P \cdot N)
        PdotN = self.polynomials_product(P,N,simplify_coeff=simplify_coeff) 
        
        #\int (N]1+] \cdot P)
        IntNdotP = self.integrate_polynomial(
                        self.polynomials_product(N1p,P,simplify_coeff=simplify_coeff),
                        simplify_coeff=simplify_coeff
                    ) 

        half_reduced_product = self.polynomials_add(
                            PdotN ,self.product_P_Coeff(IntNdotP, -1,simplify_coeff=simplify_coeff),
                            simplify_coeff=simplify_coeff
                        )
        return half_reduced_product
    
    def reduced_product(self, 
                        P:IntegralPolynomial,
                        Q:IntegralPolynomial,
                        simplify_coeff=True) -> IntegralPolynomial:
        """
        Let P and Q such that
        P_I, P_N, Q_I, Q_N are nonzero
        with lm{P}=lm{P_I} and lm{Q}=lm{Q_I}.
        Then
        reduced-product =  
            -P·Q + P_N·Q  + Q_N·P   
            + int{ (P_I]1+] · Q) + (Q_I]1+] · P) }

        or equivalently 

        reduced-product =  
            int{ (P_I]1+]  · Q_N) + (Q_I]1+]  · P_N) } + P_N · Q_N

        Moreover, the reduced-product is commutative and associative.
        """
        LM_P,_ = self.LM_LC(P)
        P_I = P.get_P_I()
        LM_Q,_ = self.LM_LC(Q)
        Q_I = Q.get_P_I()
        assert not P_I .is_zero()
        assert not Q_I .is_zero()
        LM_P_I,_ = self.LM_LC(P_I)
        LM_Q_I,_ = self.LM_LC(Q_I)
        assert LM_P == LM_P_I 
        assert LM_Q == LM_Q_I  
        P_N = P.get_P_N()
        Q_N = Q.get_P_N()

        # we need to compute redprod(P,Q)= Pn * Qn + int( Q_I[1+] PN  + P_I[1+] QN )
        P_N_dot_Q_N = self.polynomials_product(P_N,Q_N,simplify_coeff=False)

        P_I_cut_1plus = P_I.cut_P("1+")
        Q_I_cut_1plus = Q_I.cut_P("1+")

        P_I_cut_1plus_dot_Q_N = self.polynomials_product(P_I_cut_1plus,Q_N,simplify_coeff=False)
        Q_I_cut_1plus_dot_P_N = self.polynomials_product(Q_I_cut_1plus,P_N,simplify_coeff=False)


        #int(Q_I[1+] PN  + P_I[1+] QN)
        res = self.polynomials_add(
            self.integrate_polynomial(P_I_cut_1plus_dot_Q_N,simplify_coeff=False),
            self.integrate_polynomial(Q_I_cut_1plus_dot_P_N,simplify_coeff=False),
            simplify_coeff=False
        )
        reduced_product = self.polynomials_add(res,P_N_dot_Q_N, simplify_coeff=simplify_coeff)
        return reduced_product

    def polynomial_power(self, 
                        P:IntegralPolynomial,
                        n:int,
                        simplify_coeff=True) -> IntegralPolynomial:
        """
        Return: P^n
        """
        assert is_int(n) 
        assert isinstance(P,IntegralPolynomial)
        if n == 0: return IntegralPolynomial(IM(1))
        P_pow_n = P
        for _ in range(n-1):
            P_pow_n = self.polynomials_product(
                        P_pow_n,
                        P,
                        simplify_coeff=simplify_coeff
            )
        return P_pow_n
    
    def reduced_power(self, 
                      P:IntegralPolynomial, 
                      n:int,
                      simplify_coeff=True) -> IntegralPolynomial:
        """  
        Consider a nonzero integral polynomial P written as
        P = P_I + P_N 
        with P_I !=0, P_N !=0 and lm{P}=lm{P_I}.
        For any positive integer n,
        define
        reduced_power(P,n) =  
            n int{ P_I]1+] · P_N^{n-1} } + P_N^n
        """
        assert is_int(n) and n >= 1 
        if n==1: return P
        P_I = P.get_P_I()
        P_N = P.get_P_N() 
        P_I_1plus = P_I.cut_P("1+") 
        P_N_pow_n_minus_one = self.polynomial_power(
                        P_N, 
                        n-1,
                        simplify_coeff=simplify_coeff
        )

        P_N_pow_n = self.polynomials_product(
            P_N_pow_n_minus_one, 
            P_N,
            simplify_coeff=simplify_coeff
        )

        #lets compute n int{ P_I]1+] · P_N^{n-1} } + P_N^n 
        temp = self.integrate_polynomial(
                    self.polynomials_product(
                        P_I_1plus, 
                        P_N_pow_n_minus_one,
                        simplify_coeff=simplify_coeff
                     ),
                simplify_coeff=simplify_coeff
            )

        temp = self.product_P_Coeff(temp, n, simplify_coeff=simplify_coeff)

        reduced_power = self.polynomials_add(
                    temp, 
                    P_N_pow_n,
                    simplify_coeff=simplify_coeff
            )

        return reduced_power
    
    def reduction_M_by_P_simple_case(self,
                                    M: IM, 
                                    P: IntegralPolynomial
                                    ) -> IntegralPolynomial:
        return reduction_M_by_P_simple_case(self, M, P)
    
    def reduction_M_by_P_reduced_power(self,
                                       M: IM, 
                                       P: IntegralPolynomial
                                       ) -> IntegralPolynomial:
        return reduction_M_by_P_reduced_power(self, M, P)
    
    def reduction_M_by_P(self,
                         M: IM, 
                         P: IntegralPolynomial
                        ) -> IntegralPolynomial:
        return reduction_M_by_P(self,M,P)
    
    def reduce(self, 
                Q: IntegralPolynomial, 
                T: OrderedSet[IntegralPolynomial]
            ) -> IntegralPolynomial:
        return reduce(self,Q,T)
    
    def auto_reduce(self,  
                    T:OrderedSet[IntegralPolynomial]
                ) -> OrderedSet[IntegralPolynomial]:
        return auto_reduce(self,T)
    
    def critical_pairs_PI_QI(self,
                            P: IntegralPolynomial,
                            Q: IntegralPolynomial) -> IntegralPolynomial:
        return critical_pairs_PI_QI(self,P,Q)
    
    def critical_pairs_PI_QN(self,
                            P: IntegralPolynomial,
                            Q: IntegralPolynomial) -> IntegralPolynomial:
        return critical_pairs_PI_QN(self,P,Q)
    
    def critical_pairs_PN_QN(self,
                            P: IntegralPolynomial,
                            Q: IntegralPolynomial) -> IntegralPolynomial:
        return critical_pairs_PN_QN(self,P,Q)
    
    def critical_pairs(self,
                       R: OrderedSet[IntegralPolynomial]
                    ) -> OrderedSet [IntegralPolynomial]:
        return critical_pairs(self,R)
    
    def find_A_A0_G_F(self, 
                      P:IntegralPolynomial
                      ) -> tuple[IntegralPolynomial]:
        return find_A_A0_G_F(self,P)
    
    def update_exp(self, 
                T_prime: OrderedSet[IntegralPolynomial],
                E: OrderedSet[sp.Function, sp.Function, IntegralPolynomial]
                ) -> tuple:
        return update_exp(self,T_prime,E)
         
    def extend_X_with_exp_and_log(self,
                      E: set[sp.Function, sp.Function, IntegralPolynomial],
                      L: OrderedSet[sp.Function, IntegralPolynomial]
                      ) -> list:
        return extend_X_with_exp_and_log(self, E,L)

    def update_log(self, 
                T_prime: OrderedSet[IntegralPolynomial],
                L: OrderedSet[sp.Function, IntegralPolynomial]
                ) -> tuple:
        return update_log(self,T_prime,L)
    
    def integral_elimination(self,
                            F: OrderedSet[IntegralPolynomial],
                            disable_exp: bool = False,
                            disable_log: bool = False,
                            disable_critical_pairs: bool = False,
                            nb_iter: int = 0) -> tuple:
        return integral_elimination(self, 
                                F, 
                                disable_exp=disable_exp,
                                disable_log=disable_log,
                                disable_critical_pairs=disable_critical_pairs,
                                nb_iter=nb_iter)